"use strict";

/* eslint-disable array-element-newline */
module.exports = new Set([
  "break", "case", "class", "catch", "const", "continue", "debugger", "default", "delete",
  "do", "else", "export", "extends", "finally", "for", "function", "if", "import", "in", "instanceof",
  "let", "new", "return", "super", "switch", "this", "throw", "try", "typeof", "var", "void", "while", "with", "yield",

  "enum", "await",

  "package", "protected", "static", "interface", "private", "public",

  "null", "true", "false"
]);
